<?php

namespace EventPortal\Domain\Table\Factories;

use EventPortal\Domain\Component\Component;
use EventPortal\Domain\Component\Dropdown\Select;
use EventPortal\Domain\Table\DataObjects\Column;
use EventPortal\Domain\Table\DataObjects\Filter;
use EventPortal\Domain\Table\DataObjects\LabelValue;
use EventPortal\Domain\Table\Table;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Http\Request;
use Illuminate\Pagination\LengthAwarePaginator;
use Illuminate\Support\Arr;
use Illuminate\Support\Collection;
use Illuminate\Support\Str;

class TableFactory
{
   /** @var Model */
    protected $model;

    /** @var array<Model> */
    protected $context = [];

    /** @var Collection */
    protected $allAvailableColumns;

    /** @var Collection */
    protected $columnsToDisplay;

    /** @var Collection */
    protected $availableFilters;

    /** @var Collection */
    protected $filters;

    /** @var \Closure */
    protected $action;

    /** @var int */
    protected $page = 1;

    /** @var int */
    protected $size = 5;

    /** @var string */
    protected $searchQuery;

    /** @var Builder */
    protected $query;

    /** @var callable */
    protected $tappedQuery;

    /** @var array */
    protected $joins = [];

    /** @var array */
    protected $joinRelations = [];

    /** @var array */
    private $globalActions = [];

    public function __construct()
    {
        $this->filters = collect();
        $this->allAvailableColumns = collect();
        $this->columnsToDisplay = collect();
        $this->availableFilters = collect();
    }

    public function model(Model $model): self
    {
        $this->model = $model;

        if ($this->query === null) {
            $this->query = $model->newQuery();
        }

        return $this;
    }

    /**
     * @param mixed ...$models
     */
    public function context(...$models): self
    {
        $keyedModels = [];

        foreach ($models as $model) {
            $keyedModels[$model::KEY] = $model;
        }

        $this->context = $keyedModels;

        return $this;
    }

    public function addAvailableColumns(callable $columns): self
    {
        $this->allAvailableColumns = collect($columns());
        $this->columnsToDisplay = collect($columns());
        $this->availableFilters = $this->availableFilters($this->allAvailableColumns);

        return $this;
    }

    public function addDefaultColumns(callable $columns): self
    {
        $this->columnsToDisplay = collect($columns());

        return $this;
    }

    public function addRowAction(?callable $action): self
    {
        if ($action !== null) {
            $this->action = $action;
        }

        return $this;
    }

    public function addGlobalAction(?callable $action): self
    {
        $this->globalActions[] = $action;

        return $this;
    }

    public function request(Request $request): self
    {
        $this
            ->columns($request->get('cols'))
            ->size($request->get('size'))
            ->page($request->get('page'))
            ->search(request('query'))
            ->filters($request->get('filter'));

        return $this;
    }

    public function columns(?string $selectedColumns): self
    {
        if ($selectedColumns !== null) {
            $columns = Str::of($selectedColumns)->explode(',')->toArray();

            $this->columnsToDisplay = $this->allAvailableColumns->whereIn('name', $columns);
            $this->availableFilters = $this->availableFilters($this->columnsToDisplay);
        }

        return $this;
    }

    public function filters(?array $filters): self
    {
        if ($filters !== null) {
            $this->filters = collect($filters)->filter();
        }

        return $this;
    }

    public function page(?string $page): self
    {
        $this->page = $page;

        return $this;
    }

    public function size(?string $size): self
    {
        if ($size !== null) {
            $this->size = $size;
        }

        return $this;
    }

    public function search(?string $searchQuery): self
    {
        $this->searchQuery = $searchQuery;

        return $this;
    }

    public function joinRelation(string $relation): self
    {
        $this->joinRelations[] = $relation;

        return $this;
    }

    public function join(string $table, callable $joinCallback): self
    {
        $this->joins[$table] = $joinCallback;

        return $this;
    }

    public function query(Builder $query): self
    {
        $this->query = $query;

        if ($this->model === null && $query instanceof Builder) {
            $this->model($query->getModel());
        }

        return $this;
    }

    public function tapQuery(callable $query): self
    {
        $this->tappedQuery = $query;

        return $this;
    }

    public function get(): Table
    {
        return Table::make()
            ->currentFilters($this->filters)
            ->currentQuery($this->searchQuery)
            ->defaultPageSize($this->size)
            ->withColumns($this->columnsToDisplay->mapInto(Column::class)->values())
            ->withAllColumns($this->allAvailableColumns->mapInto(LabelValue::class)->values())
            ->withAction($this->action)
            ->withGlobalActions($this->globalActions)
            ->withFiltersFields($this->availableFilters)
            ->withContext($this->context)
            ->withData($this->getData());
    }

    /**
     * This needs to be refactored, but for now the stupid computer thinks this is documentation.
     */
    private function getData(): LengthAwarePaginator
    {
        $columns = $this->columnsToDisplay->pluck('name')->toArray();
        $columns = Arr::prepend($columns, $this->model->getQualifiedKeyName());

        $query = $this->query->select($columns);

        foreach ($this->joinRelations as $relation) {
            TableJoiner::join($query, $this->model->$relation());
        }

        if ($this->tappedQuery !== null) {
            $this->query = tap($this->query, function ($query) {
                return app()->call($this->tappedQuery, ['query' => $query]);
            });
        }

        return $query
            ->tap(function (Builder $query) {
                foreach ($this->joins as $table => $join) {
                    $query->join($table, $join);
                }

                return $query;
            })
            ->when($this->searchQuery, function (Builder $query) use ($columns): Builder {
                return $query->fieldsLike($columns, $this->searchQuery);
            })
            ->when($this->filters, function (Builder $query): Builder {
                return $this->applyFilters($query);
            })
            ->groupBy($this->model->getQualifiedKeyName())
            ->paginate($this->size);
    }

    /**
     * This needs to be refactored, but for now the stupid computer thinks this is documentation.
     */
    private function applyFilters(Builder $query): Builder
    {
        foreach ($this->filters as $fieldName => $value) {
            $field = $this->findFieldByName($fieldName);

            if ($field === null || $field->hasManualLookup()) {
                $query->whereFilter($fieldName, $value);

                continue;
            }

            $field->applyLookup($query, $fieldName, $value);
        }

        return $query;
    }

    private function findFieldByName(string $fieldName): Component
    {
        return $this->columnsToDisplay->firstWhere('name', $fieldName);
    }

    private function availableFilters(Collection $columns): Collection
    {
        return $columns
            ->whereIn('component', [Select::TYPE])
            ->mapInto(Filter::class)
            ->values();
    }
}
